#!/usr/bin/python
# Copyright (c) 2017 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from contextlib import closing
from contextlib import contextmanager
from email.parser import HeaderParser
import os
import shutil
import tarfile
import tempfile
import zipfile

import six
from six.moves.urllib.request import urlopen


def _download_file(url, dest_dir, dest_filename):
    """download a given url to a given destination directory and
    destination filenamee"""
    filename = os.path.join(dest_dir, dest_filename)
    with closing(urlopen(url)) as response:  # nosec
        with open(filename, 'wb') as f:
            while True:
                buf = response.read(8192)
                if not buf:
                    break
                f.write(buf)


@contextmanager
def _extract_archive_to_tempdir(archive_filename):
    """extract the given tarball or zipfile to a tempdir.
    Delete the tempdir at the end"""
    if not os.path.exists(archive_filename):
        raise Exception("Archive '%s' does not exist" % (archive_filename))

    tempdir = tempfile.mkdtemp(prefix="renderspec_")
    try:
        if tarfile.is_tarfile(archive_filename):
            with tarfile.open(archive_filename) as f:
                f.extractall(tempdir)
        elif zipfile.is_zipfile(archive_filename):
            with zipfile.ZipFile(archive_filename) as f:
                f.extractall(tempdir)
        else:
            raise Exception("Can not extract '%s'. "
                            "Not a tar or zip file" % archive_filename)
        yield tempdir
    finally:
        shutil.rmtree(tempdir)


def _find_archives(directories, basename):
    """return a list of archives in the given directories
    or an empty list if no archive(s) can be found"""
    if isinstance(directories, six.string_types):
        directories = [directories]

    return sorted(
        [os.path.join(d, f) for d in directories if d for f in os.listdir(d)
         if f.startswith(basename) and
         f.endswith(('tar.gz', 'zip', 'tar.bz2', 'xz'))],
        key=lambda x: os.stat(x).st_mtime, reverse=True)


def _find_pkg_info(directory):
    """find and return the full path to a PKG-INFO file or None if not found"""
    for root, dirs, files in os.walk(directory):
        for filename in files:
            if filename == 'PKG-INFO':
                return os.path.join(root, filename)
    # no PKG-INFO file found
    return None


def _get_version_from_pkg_info(pkg_info_filename):
    """get the version from a PKG-INFO (see pep-0314) file"""
    with open(pkg_info_filename, 'r') as f:
        parser = HeaderParser()
        headers = parser.parse(f)
        return headers.get('version')



#!/usr/bin/python
# Copyright (c) 2015 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import argparse
import os
import platform
import string
import sys

from jinja2 import Environment

import yaml

from renderspec.distloader import RenderspecLoader
from renderspec import versions
from renderspec import contextfuncs


def generate_spec(spec_style, epochs, requirements, input_template_format,
                  input_template_path, output_path):
    """generate a spec file with the given style and input template"""
    if input_template_format == 'spec.j2':
        return _renderer_input_template_format_spec(
            spec_style, epochs, requirements, input_template_path,
            output_path)
    else:
        raise Exception('Unknown input-template-format "%s"' %
                        input_template_format)


def _renderer_input_template_format_spec(spec_style, epochs, requirements,
                                         input_template_path, output_path):
    """render a 'traditional' .spec.j2 template into a .spec file"""
    env = Environment(loader=RenderspecLoader(
        template_fn=input_template_path),
        trim_blocks=True)

    contextfuncs.env_register_filters_and_globals(env)

    template_name = '.spec'
    if spec_style in env.loader.list_templates():
        template_name = spec_style
    template = env.get_template(template_name)
    input_template_dir = os.path.dirname(os.path.abspath(input_template_path))
    if output_path:
        output_dir = os.path.dirname(
            os.path.abspath(output_path))
    else:
        output_dir = None
    return template.render(spec_style=spec_style, epochs=epochs,
                           requirements=requirements,
                           input_template_dir=input_template_dir,
                           output_dir=output_dir)


def _is_fedora(distname):
    """detect Fedora-based distro (e.g Fedora, CentOS, RHEL)"""
    distname = distname.lower()
    for x in ["fedora", "centos", "red hat"]:
        if x in distname:
            return True
    return False


def _get_default_distro():
    distname, _, _ = platform.linux_distribution()

    # newer distros only have /etc/os-release and then platform doesn't work
    # anymore and upstream does not want to fix it:
    # https://bugs.python.org/issue1322
    if not distname and 'Linux' in platform.system():
        try:
            with open('/etc/os-release', 'r') as lsb_release:
                for l in lsb_release:
                    if l.startswith('ID_LIKE='):
                        distname = l.partition('=')[2].strip(
                            string.punctuation + string.whitespace)
                        break
        except OSError:
            print('WARN: Unable to determine Linux distribution')

    if "suse" in distname.lower():
        return "suse"
    elif _is_fedora(distname):
        return "fedora"
    else:
        return "unknown"


def _get_default_template():
    fns = [f for f in os.listdir('.')
           if os.path.isfile(f) and f.endswith('.spec.j2')]
    if not fns:
        return None, ("No *.spec.j2 templates found. "
                      "See `renderspec -h` for usage.")
    elif len(fns) > 1:
        return None, ("Multiple *.spec.j2 templates found, "
                      "please specify one.\n"
                      "See `renderspec -h` for usage.")
    else:
        return fns[0], None


def _get_epochs(filename):
    """get a dictionary with pkg-name->epoch mapping"""
    epochs = {}
    if filename is not None:
        with open(filename, 'r') as f:
            data = yaml.safe_load(f.read())
            epochs.update(data['epochs'])
    return epochs


def _get_requirements(filenames):
    """get a dictionary with pkg-name->min-version mapping"""
    reqs = {}
    for filename in filenames:
        with open(filename, 'r') as f:
            reqs.update(versions.get_requirements(f.readlines()))
    return reqs


def process_args():
    distro = _get_default_distro()
    parser = argparse.ArgumentParser(
        description="Convert a .spec.j2 template into a .spec")
    parser.add_argument("-o", "--output",
                        help="output filename or '-' for stdout. "
                        "default: autodetect")
    parser.add_argument("--spec-style", help="distro style you want to use. "
                        "default: %s" % (distro), default=distro,
                        choices=['suse', 'fedora'])
    parser.add_argument("--epochs", help="yaml file with epochs listed.")
    parser.add_argument("input-template", nargs='?',
                        help="specfile jinja2 template to render. "
                        "default: *.spec.j2")
    parser.add_argument("-f", "--input-template-format", help="Format of the "
                        "input-template file. default: %(default)s",
                        default="spec.j2", choices=["spec.j2"])
    parser.add_argument("--requirements", help="file(s) which contain "
                        "PEP0508 compatible requirement lines. Last mentioned "
                        "file has highest priority. default: %(default)s",
                        action='append', default=[])

    return vars(parser.parse_args())


def main():
    args = process_args()

    # autodetect input/output fns if possible
    input_template = args['input-template']
    if not input_template:
        input_template, errmsg = _get_default_template()
        if not input_template:
            print(errmsg)
            return 1
    output_filename = args['output']
    if not output_filename:
        if not input_template.endswith('.spec.j2'):
            print("Failed to autodetect output file name. "
                  "Please specify using `-o/--output`.")
            return 2
        output_filename, _, _ = input_template.rpartition('.')

    try:
        epochs = _get_epochs(args['epochs'])
        requirements = _get_requirements(args['requirements'])
    except IOError as e:
        print(e)
        return 3

    if output_filename and output_filename != '-':
        output_path = os.path.abspath(output_filename)
    else:
        output_path = None

    spec = generate_spec(args['spec_style'], epochs, requirements,
                         args['input_template_format'], input_template,
                         output_path)
    if output_path:
        print("Rendering: %s -> %s" % (input_template, output_path))
        with open(output_path, "w") as o:
            o.write(spec)
    else:
        print(spec)
    return 0


if __name__ == '__main__':
    sys.exit(main())



#!/usr/bin/python
# Copyright (c) 2016 Red Hat
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import jinja2
from jinja2.loaders import TemplateNotFound
from jinja2.utils import open_if_exists
import os


def get_dist_templates_path():
    return os.path.join(os.path.dirname(__file__), 'dist-templates')


class RenderspecLoader(jinja2.BaseLoader):
    """A special template loader which allows rendering supplied .spec template
    with distro specific blocks maintained as part of renderspec.

    '.spec' returns the spec template (which you need to supply during init)
    while other strings map to corresponding child templates included
    in renderspec which simply extend the '.spec' template.
    """
    base_ref = '.spec'
    template_postfix = '.spec.j2'

    def __init__(self, template_fn, encoding='utf-8'):
        self.base_fn = template_fn
        self.encoding = encoding
        self.disttemp_path = get_dist_templates_path()

    def get_source(self, environment, template):
        if template == self.base_ref:
            fn = self.base_fn
        else:
            fn = os.path.join(self.disttemp_path,
                              template + self.template_postfix)

        f = open_if_exists(fn)
        if not f:
            return TemplateNotFound(template)
        try:
            contents = f.read().decode(self.encoding)
        finally:
            f.close()

        mtime = os.path.getmtime(self.base_fn)

        def uptodate():
            try:
                return os.path.getmtime(self.base_fn) == mtime
            except OSError:
                return False

        return contents, fn, uptodate

    def list_templates(self):
        found = set([self.base_ref])
        walk_dir = os.walk(self.disttemp_path)
        for _, _, filenames in walk_dir:
            for fn in filenames:
                if fn.endswith(self.template_postfix):
                    template = fn[:-len(self.template_postfix)]
                    found.add(template)
        return sorted(found)



# Copyright (c) 2017 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

import os

from jinja2 import contextfilter
from jinja2 import contextfunction
from jinja2.exceptions import TemplateRuntimeError
from packaging.version import parse
import pymod2pkg

from renderspec import utils


# a variable that needs to be set for some functions in the context
CONTEXT_VAR_PYPI_NAME = "pypi_name"
CONTEXT_VAR_UPSTREAM_VERSION = "upstream_version"
CONTEXT_VAR_RPM_RELEASE = "rpm_release"


def _context_check_variable(context, var_name, needed_by):
    """check that the context has a given variable"""
    if var_name not in context.vars:
        raise TemplateRuntimeError("Variable '%s' not available in context but"
                                   " needed for '%s'" % (var_name, needed_by))


def _context_fetch_source(context, url):
    """fetch the given url into the output_dir and return the url"""
    if context['output_dir']:
        filename = os.path.basename(url)
        utils._download_file(url, context['output_dir'], filename)
    return url


def _context_url_pypi(context):
    """return the full sdist pypi url"""
    # we need the pypi_name and the upstream_version variables to construct
    # the full url
    _context_check_variable(context, CONTEXT_VAR_PYPI_NAME,
                            'pypi_name')

    _context_check_variable(context, CONTEXT_VAR_UPSTREAM_VERSION,
                            'upstream_version')
    name = context.vars[CONTEXT_VAR_PYPI_NAME]
    version = context.vars[CONTEXT_VAR_UPSTREAM_VERSION]
    return 'https://files.pythonhosted.org/packages/source/' \
        '%s/%s/%s-%s.tar.gz' % (name[0], name, name, version)


def _context_upstream_version(context, pkg_version=None):
    """return the version which should be set to the 'upstream_version'
    variable in the jinja context"""
    if pkg_version:
        return pkg_version
    else:
        # try to auto-detect the version - for that we need the pypi name
        _context_check_variable(context, CONTEXT_VAR_PYPI_NAME,
                                'upstream_version')
        pypi_name = context.vars[CONTEXT_VAR_PYPI_NAME]

        # look for archives in:
        # 1) the output_dir
        # 2) the dir where the input template (.spec.j2) comes from
        # 3) the current working dir
        archives = utils._find_archives([context['output_dir'],
                                         context['input_template_dir'],
                                         '.'], pypi_name)
        for archive in archives:
            with utils._extract_archive_to_tempdir(archive) as tmpdir:
                pkg_info_file = utils._find_pkg_info(tmpdir)
                if pkg_info_file:
                    return utils._get_version_from_pkg_info(pkg_info_file)
        # unable to autodetect the version
        raise TemplateRuntimeError("Can not autodetect 'upstream_version' from"
                                   " the following archives: '%s'" % (
                                       ', '.join(archives)))


def _context_py2rpmversion(context):
    """get a python PEP0440 compatible version and translate it to an RPM
    version"""
    # the context needs a variable set via {% set upstream_version = 'ver' %}
    _context_check_variable(context, CONTEXT_VAR_UPSTREAM_VERSION,
                            'py2rpmversion')
    version = context.vars[CONTEXT_VAR_UPSTREAM_VERSION]
    v_python = parse(version)
    # fedora does not allow '~' in versions but uses a combination of Version
    # and Release
    # https://fedoraproject.org/wiki/Packaging:Versioning\#Pre-Release_packages
    if context['spec_style'] == 'fedora':
        if len(v_python._version.release) >= 4:
            return "%d.%d.%d" % (v_python._version.release[0:3])
        else:
            return v_python.base_version
    else:
        v_rpm = v_python.public
        if v_python.is_prerelease:
            # we need to add the 'x' in front of alpha/beta releases because
            # in the python world, "1.1a10" > "1.1.dev10"
            # but in the rpm world, "1.1~a10" < "1.1~dev10"
            v_rpm = v_rpm.replace('a', '~xalpha')
            v_rpm = v_rpm.replace('b', '~xbeta')
            v_rpm = v_rpm.replace('rc', '~xrc')
            v_rpm = v_rpm.replace('.dev', '~dev')
        return v_rpm


def _context_py2rpmrelease(context):
    if context['spec_style'] == 'fedora':
        # the context needs a var set via {% set upstream_version = 'ver' %}
        _context_check_variable(context, CONTEXT_VAR_UPSTREAM_VERSION,
                                'py2rpmrelease')
        # the context needs a var set via {% set rpm_release = 'ver' %}
        _context_check_variable(context, CONTEXT_VAR_RPM_RELEASE,
                                'py2rpmrelease')
        upstream_version = context.vars[CONTEXT_VAR_UPSTREAM_VERSION]
        rpm_release = context.vars[CONTEXT_VAR_RPM_RELEASE]
        v_python = parse(upstream_version)
        if v_python.is_prerelease:
            _, alphatag = v_python.public.split(v_python.base_version)
            return '0.{}.{}%{{?dist}}'.format(rpm_release,
                                              alphatag.lstrip('.'))
        else:
            return '{}%{{?dist}}'.format(rpm_release)
    else:
        # SUSE uses just '0'. The OpenBuildService handles the Release tag
        return '0'


def _context_epoch(context, pkg_name):
    """get the epoch (or 0 if unknown) for the given pkg name"""
    return context['epochs'].get(pkg_name, 0)


def _pymod2pkg_translate(pkg_name, spec_style, py_versions):
    """translate a given package name for a single or multiple py versions"""
    if py_versions and not isinstance(py_versions, (list, tuple)):
        py_versions = [py_versions]
    kwargs = {'py_vers': py_versions} if py_versions else {}
    translations = pymod2pkg.module2package(pkg_name, spec_style, **kwargs)
    # we want always return a list but module2package() might return a string
    if not isinstance(translations, (list, tuple)):
        translations = [translations]
    return translations


def _context_py2name(context, pkg_name=None, pkg_version=None,
                     py_versions=None):
    """
    context: a Jinja2 context
    pkg_name: usually the pypi-name. If None, it tries to get the name
              from the context variable called 'py2name'
    pkg_version: Deprecated and unused
    py_versions: the version pymod2pkg should return. Can be currently 'py',
                 'py2' and 'py3' or a combination of those in a list
    """
    if not pkg_name:
        # if the name is not given, try to get the name from the context
        _context_check_variable(context, CONTEXT_VAR_PYPI_NAME,
                                'py2name')
        pkg_name = context.vars[CONTEXT_VAR_PYPI_NAME]
    # return always a string to be backwards compat
    return ' '.join(_pymod2pkg_translate(
        pkg_name, context['spec_style'], py_versions))


def _context_py2pkg(context, pkg_name, pkg_version=None, py_versions=None):
    """generate a distro specific package name with optional version tuple."""
    name_list = _pymod2pkg_translate(pkg_name, context['spec_style'],
                                     py_versions)

    # if no pkg_version is given, look in the requirements and set one
    if not pkg_version:
        if pkg_name in context['requirements']:
            pkg_version = ('>=', context['requirements'][pkg_name])

    # pkg_version is a tuple with comparator and number, i.e. "('>=', '1.2.3')"
    if pkg_version:
        # epoch handling
        if pkg_name in context['epochs'].keys():
            epoch = '%s:' % context['epochs'][pkg_name]
        else:
            epoch = ''
        v_comparator, v_number = pkg_version
        v_str = ' %s %s%s' % (v_comparator, epoch, v_number)
    else:
        v_str = ''

    return ' '.join(['%s%s' % (name, v_str) for name in name_list])


def _context_license_spdx(context, value):
    """convert a given known spdx license to another one"""
    # more values can be taken from from https://github.com/hughsie/\
    #    appstream-glib/blob/master/libappstream-builder/asb-package-rpm.c#L76
    mapping = {
        "Apache-1.1": "ASL 1.1",
        "Apache-2.0": "ASL 2.0",
        "BSD-3-Clause": "BSD",
        "GPL-1.0+": "GPL+",
        "GPL-2.0": "GPLv2",
        "GPL-2.0+": "GPLv2+",
        "GPL-3.0": "GPLv3",
        "GPL-3.0+": "GPLv3+",
        "LGPL-2.1": "LGPLv2.1",
        "LGPL-2.1+": "LGPLv2+",
        "LGPL-2.0": "LGPLv2 with exceptions",
        "LGPL-2.0+": "LGPLv2+ with exceptions",
        "LGPL-3.0": "LGPLv3",
        "LGPL-3.0+": "LGPLv3+",
        "MIT": "MIT with advertising",
        "MPL-1.0": "MPLv1.0",
        "MPL-1.1": "MPLv1.1",
        "MPL-2.0": "MPLv2.0",
        "OFL-1.1": "OFL",
        "Python-2.0": "Python",
    }

    if context['spec_style'] == 'fedora':
        return mapping[value]
    else:
        # just use the spdx license name
        return value


###############
# jinja2 filter
###############
@contextfilter
def _filter_epoch(context, value):
    return _context_epoch(context, value)


@contextfilter
def _filter_basename(context, value):
    return os.path.basename(value)


################
# jinja2 globals
################
@contextfunction
def _globals_py2pkg(context, pkg_name, pkg_version=None, py_versions=None):
    return _context_py2pkg(context, pkg_name, pkg_version, py_versions)


@contextfunction
def _globals_fetch_source(context, url):
    return _context_fetch_source(context, url)


@contextfunction
def _globals_url_pypi(context):
    return _context_url_pypi(context)


@contextfunction
def _globals_upstream_version(context, pkg_version=None):
    return _context_upstream_version(context, pkg_version)


@contextfunction
def _globals_py2rpmversion(context):
    return _context_py2rpmversion(context)


@contextfunction
def _globals_py2rpmrelease(context):
    return _context_py2rpmrelease(context)


@contextfunction
def _globals_epoch(context, value):
    return _context_epoch(context, value)


@contextfunction
def _globals_license_spdx(context, value):
    return _context_license_spdx(context, value)


@contextfunction
def _globals_py2name(context, value=None, py_versions=None):
    return _context_py2name(context, value, py_versions=py_versions)


def env_register_filters_and_globals(env):
    """register all the jinja2 filters we want in the environment"""
    env.filters['epoch'] = _filter_epoch
    env.filters['basename'] = _filter_basename
    env.globals['py2rpmversion'] = _globals_py2rpmversion
    env.globals['py2rpmrelease'] = _globals_py2rpmrelease
    env.globals['py2pkg'] = _globals_py2pkg
    env.globals['py2name'] = _globals_py2name
    env.globals['epoch'] = _globals_epoch
    env.globals['license'] = _globals_license_spdx
    env.globals['upstream_version'] = _globals_upstream_version
    env.globals['fetch_source'] = _globals_fetch_source
    env.globals['url_pypi'] = _globals_url_pypi



#!/usr/bin/python
# Copyright (c) 2016 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


try:
    import unittest2 as unittest
except ImportError:
    import unittest

from ddt import data, ddt, unpack

from jinja2 import Environment
from jinja2.exceptions import TemplateRuntimeError

from mock import Mock, patch
import os
import renderspec
import renderspec.contextfuncs
import renderspec.utils
import renderspec.versions
import shutil
import tempfile
import time


@ddt
class RenderspecContextFunctionTests(unittest.TestCase):
    """test functions which do some calculation based on the context"""
    def test_context_license_spdx(self):
        self.assertEqual(
            renderspec.contextfuncs._context_license_spdx(
                {'spec_style': 'suse'}, 'Apache-2.0'),
            'Apache-2.0'
        )
        self.assertEqual(
            renderspec.contextfuncs._context_license_spdx(
                {'spec_style': 'fedora'}, 'Apache-2.0'),
            'ASL 2.0'
        )

    @data(
        # without version
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         'oslo.config', None, None, 'python-oslo.config'),
        ({'spec_style': 'fedora', 'epochs': {}, 'requirements': {}},
         'oslo.config', None, None, 'python-oslo-config'),
        # without version, multiple python versions
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         'oslo.config', None, ('py', 'py3'),
         'python-oslo.config python3-oslo.config'),
        # with version
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         'oslo.config', ('>=', '1.2.3'), None, 'python-oslo.config >= 1.2.3'),
        ({'spec_style': 'fedora', 'epochs': {}, 'requirements': {}},
         'oslo.config', ('==', '1.2.3~a0'), None,
         'python-oslo-config == 1.2.3~a0'),
        # with version, with epoch
        ({'spec_style': 'suse', 'epochs': {'oslo.config': 4},
          'requirements': {}},
         'oslo.config', ('>=', '1.2.3'), None,
         'python-oslo.config >= 4:1.2.3'),
        # without version, with epoch
        ({'spec_style': 'suse', 'epochs': {'oslo.config': 4},
          'requirements': {}},
         'oslo.config', None, None, 'python-oslo.config'),
        # with version, with requirements
        ({'spec_style': 'suse', 'epochs': {},
          'requirements': {'oslo.config' '1.2.3'}},
         'oslo.config', ('>=', '4.5.6'), None, 'python-oslo.config >= 4.5.6'),
        # without version, with requirements
        ({'spec_style': 'suse', 'epochs': {},
          'requirements': {'oslo.config': '1.2.3'}},
         'oslo.config', None, None, 'python-oslo.config >= 1.2.3'),
        # without version, with requirements, with epoch
        ({'spec_style': 'suse', 'epochs': {'oslo.config': 4},
          'requirements': {'oslo.config': '1.2.3'}},
         'oslo.config', None, None, 'python-oslo.config >= 4:1.2.3'),
        # with version, with requirements, with epoch
        ({'spec_style': 'suse', 'epochs': {'oslo.config': 4},
          'requirements': {'oslo.config' '1.2.3'}},
         'oslo.config', ('>=', '4.5.6'), None,
         'python-oslo.config >= 4:4.5.6'),
        # with version, with requirements, with epoch, python2
        ({'spec_style': 'suse', 'epochs': {'oslo.config': 4},
          'requirements': {'oslo.config' '1.2.3'}},
         'oslo.config', ('>=', '4.5.6'), 'py2',
         'python2-oslo.config >= 4:4.5.6'),
        # with version, with requirements, with epoch, python3
        ({'spec_style': 'suse', 'epochs': {'oslo.config': 4},
          'requirements': {'oslo.config' '1.2.3'}},
         'oslo.config', ('>=', '4.5.6'), 'py3',
         'python3-oslo.config >= 4:4.5.6'),
        # with version, with requirements, with epoch, python2 and python3
        ({'spec_style': 'suse', 'epochs': {'oslo.config': 4},
          'requirements': {'oslo.config' '1.2.3'}},
         'oslo.config', ('>=', '4.5.6'), ['py2', 'py3'],
         'python2-oslo.config >= 4:4.5.6 python3-oslo.config >= 4:4.5.6'),

    )
    @unpack
    def test_context_py2pkg(self, context, pkg_name, pkg_version,
                            py_versions, expected_result):
        self.assertEqual(
            renderspec.contextfuncs._context_py2pkg(
                context, pkg_name, pkg_version, py_versions),
            expected_result)

    def test_context_epoch_without_epochs(self):
        self.assertEqual(
            renderspec.contextfuncs._context_epoch(
                {'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
                'oslo.config'), 0)

    def test_context_epoch_with_epochs(self):
        self.assertEqual(
            renderspec.contextfuncs._context_epoch(
                {'spec_style': 'suse', 'epochs': {'oslo.config': 4},
                 'requirements': {}}, 'oslo.config'), 4)

    def test_context_upstream_version(self):
        context = {'spec_style': 'suse', 'epochs': {},
                   'requirements': {}}
        self.assertEqual(renderspec.contextfuncs._context_upstream_version(
                context, '1.2.0'), '1.2.0')

    @data(
        # no output_dir -> download_files should not be called
        (None, 0),
        # output_dir defined -> download_files should be called once
        ('.', 1)
    )
    @unpack
    def test_context_fetch_source_no_output_dir(self, output_dir,
                                                expected_calls):
        context = {'spec_style': 'suse', 'epochs': {},
                   'requirements': {}, 'output_dir': output_dir}
        url = 'http://foo/bar'
        with patch('renderspec.utils._download_file') as m:
            self.assertEqual(renderspec.contextfuncs._context_fetch_source(
                context, url), url)
            self.assertEqual(m.call_count, expected_calls)


@ddt
class RenderspecTemplateFunctionTests(unittest.TestCase):
    def setUp(self):
        """create a Jinja2 environment and register the standard filters"""
        self.env = Environment()
        renderspec.contextfuncs.env_register_filters_and_globals(self.env)

    @data(
        ("{{ 'http://foo/bar'|basename }}", "bar")
    )
    @unpack
    def test_render_func_basename(self, input, expected):
        template = self.env.from_string(input)
        self.assertEqual(
            template.render(spec_style='suse', epochs={}, requirements={}),
            expected)

    def test_render_func_license_spdx(self):
        template = self.env.from_string(
            "{{ license('Apache-2.0') }}")
        self.assertEqual(
            template.render(spec_style='fedora', epochs={}, requirements={}),
            'ASL 2.0')

    @data(
        # plain
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         "{{ py2pkg('requests') }}", "python-requests"),
        # plain, with multiple py_versions
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         "{{ py2pkg('requests', py_versions=['py2', 'py3']) }}",
         "python2-requests python3-requests"),
        # with version
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         "{{ py2pkg('requests', ('>=', '2.8.1')) }}",
         "python-requests >= 2.8.1"),
        # with version, with epoch
        ({'spec_style': 'suse', 'epochs': {'requests': 4}, 'requirements': {}},
         "{{ py2pkg('requests', ('>=', '2.8.1')) }}",
         "python-requests >= 4:2.8.1"),
        # with version, with epoch, with requirements
        ({'spec_style': 'suse', 'epochs': {'requests': 4},
          'requirements': {'requests': '1.2.3'}},
         "{{ py2pkg('requests', ('>=', '2.8.1')) }}",
         "python-requests >= 4:2.8.1"),
        # without version, with epoch, with requirements
        ({'spec_style': 'suse', 'epochs': {'requests': 4},
          'requirements': {'requests': '1.2.3'}},
         "{{ py2pkg('requests') }}",
         "python-requests >= 4:1.2.3"),
        # without version, with epoch, with requirements, with py_versions
        ({'spec_style': 'suse', 'epochs': {'requests': 4},
          'requirements': {'requests': '1.2.3'}},
         "{{ py2pkg('requests', py_versions=['py2']) }}",
         "python2-requests >= 4:1.2.3"),
    )
    @unpack
    def test_render_func_py2pkg(self, context, string, expected_result):
        template = self.env.from_string(string)
        self.assertEqual(
            template.render(**context),
            expected_result)

    def test_render_func_epoch_without_epochs(self):
        template = self.env.from_string(
            "Epoch: {{ epoch('requests') }}")
        self.assertEqual(
            template.render(spec_style='suse', epochs={}, requirements={}),
            'Epoch: 0')

    def test_render_func_epoch_with_epochs(self):
        template = self.env.from_string(
            "Epoch: {{ epoch('requests') }}")
        self.assertEqual(
            template.render(spec_style='suse', epochs={'requests': 1},
                            requirements={}),
            'Epoch: 1')

    @data(
        # plain name
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         "{{ py2name('requests') }}",
         "python-requests"),
        # name with epoch
        ({'spec_style': 'suse', 'epochs': {'requests': 4}, 'requirements': {}},
         "{{ py2name('requests') }}",
         "python-requests"),
        # name with epoch, with requirements
        ({'spec_style': 'suse', 'epochs': {'requests': 4},
          'requirements': {'requests': '1.4.0'}},
         "{{ py2name('requests') }}",
         "python-requests"),
        # with pypi_name context var
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         "{% set pypi_name = 'oslo.messaging' %}{{ py2name() }}",
         "python-oslo.messaging"),
        # with pypi_name context var and explicit parameter
        ({'spec_style': 'suse', 'epochs': {}, 'requirements': {}},
         "{% set pypi_name = 'oslo.messaging' %}{{ py2name('requests') }}",
         "python-requests"),
    )
    @unpack
    def test_render_func_py2name(self, context, string, expected_result):
        """test the template context function called 'py2name()'"""
        template = self.env.from_string(string)
        self.assertEqual(
            template.render(**context),
            expected_result)

    def test_render_func_py2name_raise(self):
        """py2name() called without parameter but no pypi_name context
        variable set"""
        template = self.env.from_string(
            "{{ py2name() }}")
        with self.assertRaises(TemplateRuntimeError):
            template.render(
                **{'spec_style': 'suse', 'epochs': {}, 'requirements': {}}
            )

    @data(
        ('suse', '1.1.0', '1.1.0'),
        ('suse', '1.1.0.post2', '1.1.0.post2'),
        ('suse', '1.1.0dev10', '1.1.0~dev10'),
        ('suse', '1.1.0a10', '1.1.0~xalpha10'),
        ('suse', '1.1.0a10dev5', '1.1.0~xalpha10~dev5'),
        ('suse', '1.1.0b10', '1.1.0~xbeta10'),
        ('suse', '1.1.0rc2', '1.1.0~xrc2'),
        ('suse', '1.1.0rc2dev2', '1.1.0~xrc2~dev2'),
        ('fedora', '1.1.0', '1.1.0'),
        ('fedora', '1.1.0b10', '1.1.0'),
        ('fedora', '1.1.0rc2dev2', '1.1.0'),
        ('fedora', '11.0.0.0b1', '11.0.0'),
    )
    @unpack
    def test_render_func_py2rpmversion(self, style, py_ver, rpm_ver):
        context = {'spec_style': style, 'epochs': {}, 'requirements': {}}
        # need to escape '{' and '}' here
        s = "{{% set upstream_version = '{}' %}}{{{{ py2rpmversion() }}}}"\
            .format(py_ver)
        template = self.env.from_string(s)
        self.assertEqual(
            template.render(**context),
            rpm_ver)

    @data(
        ('suse', '1.1.0', '1', '0'),
        ('suse', '1.1.0.post2', '1', '0'),
        ('suse', '1.1.0dev10', '2', '0'),
        ('fedora', '1.1.0', '1', '1%{?dist}'),
        # ('fedora', '1.1.0.post2', '1', 'FIXME'),
        ('fedora', '1.1.0dev10', '1', '0.1.dev10%{?dist}'),
        ('fedora', '1.1.0a10', '1', '0.1.a10%{?dist}'),
        ('fedora', '1.1.0a10dev5', '1', '0.1.a10.dev5%{?dist}'),
        ('fedora', '1.1.0b10', '1', '0.1.b10%{?dist}'),
        ('fedora', '1.1.0rc2', '5', '0.5.rc2%{?dist}'),
        ('fedora', '1.1.0rc2dev2', '1', '0.1.rc2.dev2%{?dist}'),
        ('fedora', '11.0.0.0b1', '1', '0.1.b1%{?dist}'),
    )
    @unpack
    def test_render_func_py2rpmrelease(self, style, upstream_ver, rpm_release,
                                       rpm_release_expected):
        context = {'spec_style': style, 'epochs': {}, 'requirements': {}}
        # need to escape '{' and '}' here
        s = "{{% set upstream_version = '{}' %}}" \
            "{{% set rpm_release = '{}' %}}" \
            "{{{{ py2rpmrelease() }}}}".format(upstream_ver, rpm_release)
        template = self.env.from_string(s)
        self.assertEqual(
            template.render(**context),
            rpm_release_expected)

    def test_render_func_url_pypi(self):
        context = {'spec_style': 'suse', 'epochs': {}, 'requirements': {}}
        # need to escape '{' and '}' here
        s = "{% set upstream_version = '3.20.0' %}" \
            "{% set pypi_name = 'oslo.concurrency' %}" \
            "{{ url_pypi() }}"
        template = self.env.from_string(s)
        self.assertEqual(
            template.render(**context),
            "https://files.pythonhosted.org/packages/source/o/"
            "oslo.concurrency/oslo.concurrency-3.20.0.tar.gz")


class RenderspecVersionsTests(unittest.TestCase):
    def test_without_version(self):
        requires = renderspec.versions.get_requirements(
            ['# a comment', '', '   ', 'pyasn1  # BSD', 'Paste'])
        self.assertEqual(requires, {})

    def test_with_single_version(self):
        requires = renderspec.versions.get_requirements(
            ['paramiko>=1.16.0  # LGPL'])
        self.assertEqual(requires, {'paramiko': '1.16.0'})

    def test_with_multiple_versions(self):
        requires = renderspec.versions.get_requirements(
            ['sphinx>=1.1.2,!=1.2.0,!=1.3b1,<1.3  # BSD'])
        self.assertEqual(requires, {'sphinx': '1.1.2'})

    def test_lexical_version(self):
        requires = renderspec.versions.get_requirements(
            ['django>=1.8,<1.10  # FOO BAR'])
        self.assertEqual(requires, {'django': '1.8'})

    def test_with_multiple_versions_and_invalid_lowest(self):
        requires = renderspec.versions.get_requirements(
            ['sphinx>=1.1.2,!=1.1.0,!=1.3b1,<1.3  # BSD'])
        self.assertEqual(requires, {'sphinx': '1.1.2'})

    def test_with_single_marker(self):
        requires = renderspec.versions.get_requirements(
            ["pywin32>=1.0;sys_platform=='win32'  # PSF"])
        self.assertEqual(requires, {})

    def test_with_multiple_markers(self):
        requires = renderspec.versions.get_requirements(
            ["""pyinotify>=0.9.6;sys_platform!='win32' and \
            sys_platform!='darwin' and sys_platform!='sunos5' # MIT"""])
        self.assertEqual(requires, {'pyinotify': '0.9.6'})


@ddt
class RenderspecCommonTests(unittest.TestCase):
    def test__get_requirements_single_file(self):
        tmpdir = tempfile.mkdtemp(prefix='renderspec-test_')
        try:
            f1 = os.path.join(tmpdir, 'f1')
            with open(f1, 'w+') as f:
                f.write('paramiko>=1.16.0\n'
                        'pyinotify>=0.9.6')
            self.assertEqual(
                renderspec._get_requirements([f1]),
                {'paramiko': '1.16.0', 'pyinotify': '0.9.6'})
        finally:
            shutil.rmtree(tmpdir)

    def test__get_requirements_multiple_files(self):
        tmpdir = tempfile.mkdtemp(prefix='renderspec-test_')
        try:
            f1 = os.path.join(tmpdir, 'f1')
            f2 = os.path.join(tmpdir, 'f2')
            with open(f1, 'w+') as f:
                f.write('paramiko>=1.17.0  # LGPL')
            with open(f2, 'w+') as f:
                f.write('paramiko>=1.16.0  # LGPL')
            # we expect the second file was used (because mentioned last)
            self.assertEqual(renderspec._get_requirements([f1, f2]),
                             {'paramiko': '1.16.0'})
        finally:
            shutil.rmtree(tmpdir)

    @data(
        ("{{ py2pkg('requests') }}", "suse", {}, {}, "python-requests"),
        ("{{ py2pkg('requests') }}", "fedora", {}, {}, "python-requests"),
        ("{{ py2pkg('requests') }}", "suse", {}, {"requests": '1.1.0'},
         "python-requests >= 1.1.0"),
    )
    @unpack
    def test_generate_spec(self, template, style, epochs, requirements,
                           expected_result):
        tmpdir = tempfile.mkdtemp(prefix='renderspec-test_')
        try:
            f1 = os.path.join(tmpdir, 'test.spec.j2')
            with open(f1, 'w+') as f:
                f.write(template)
            rendered = renderspec.generate_spec(
                style, epochs, requirements, 'spec.j2', f1, None)
            self.assertEqual(rendered, expected_result)
        finally:
            shutil.rmtree(tmpdir)


class RenderspecDistroDetection(unittest.TestCase):
    def test_is_fedora(self):
        self.assertTrue(renderspec._is_fedora("CentOS Linux"))
        self.assertTrue(renderspec._is_fedora("Fedora"))
        self.assertTrue(renderspec._is_fedora("Red Hat Enterprise Linux 7.2"))
        self.assertFalse(renderspec._is_fedora("SUSE Linux Enterprise Server"))

    def test_get_default_distro(self):
        import platform
        platform.linux_distribution = Mock(
            return_value=('SUSE Linux Enterprise Server ', '12', 'x86_64'))
        self.assertEqual(renderspec._get_default_distro(),
                         "suse")
        platform.linux_distribution = Mock(
            return_value=('Fedora', '24', 'x86_64'))
        self.assertEqual(renderspec._get_default_distro(), "fedora")
        platform.linux_distribution = Mock(
            return_value=('Red Hat Enterprise Linux Server', '7.3', 'x86_64'))
        self.assertEqual(renderspec._get_default_distro(), "fedora")
        platform.linux_distribution = Mock(
            return_value=('CentOS Linux', '7.3', 'x86_64'))
        self.assertEqual(renderspec._get_default_distro(), "fedora")
        platform.linux_distribution = Mock(
            return_value=('debian', '7.0', 'x86_64'))
        self.assertEqual(renderspec._get_default_distro(), "unknown")


class RenderspecDistTeamplatesTests(unittest.TestCase):
    @patch('renderspec.distloader.get_dist_templates_path')
    def test_dist_templates(self, mock_dt_path):
        base_txt = ('Line before block\n'
                    '{% block footest %}{% endblock %}\n'
                    'Line after block\n')
        dt_txt = ('{% extends ".spec" %}'
                  '{% block footest %}'
                  'foo block\n'
                  'macro: {{ py2pkg("test") }}\n'
                  '{% endblock %}')
        expected_out = ('Line before block\n'
                        'foo block\n'
                        'macro: python-test\n'
                        'Line after block')
        tmpdir = tempfile.mkdtemp(prefix='renderspec-test_')
        try:
            # create .spec template
            base_path = os.path.join(tmpdir, 'foo.spec.j2')
            with open(base_path, 'w+') as f:
                f.write(base_txt)
            # create custom dist template
            dt_dir = os.path.join(tmpdir, 'dist-templates')
            os.mkdir(dt_dir)
            dt_path = os.path.join(dt_dir, 'loldistro.spec.j2')
            with open(dt_path, 'w+') as f:
                f.write(dt_txt)
            # mock this to use testing dist-tempaltes folder
            mock_dt_path.return_value = dt_dir

            out = renderspec.generate_spec('loldistro', {}, {}, 'spec.j2',
                                           base_path, None)
            self.assertEqual(out, expected_out)
        finally:
            shutil.rmtree(tmpdir)


@ddt
class RenderspecUtilsTests(unittest.TestCase):
    def _write_pkg_info(self, destdir, version='5.10.0'):
        """write a PKG-INFO file into destdir"""
        f1 = os.path.join(destdir, 'PKG-INFO')
        with open(f1, 'w+') as f:
            f.write('Metadata-Version: 1.1\n'
                    'Name: oslo.messaging\n'
                    'Version: %s' % (version))

    def test__extract_archive_to_tempdir_no_file(self):
        with self.assertRaises(Exception) as e_info:
            with renderspec.utils._extract_archive_to_tempdir("foobar"):
                self.assertIn("foobar", str(e_info))

    def test__find_pkg_info(self):
        tmpdir = tempfile.mkdtemp(prefix='renderspec-test_')
        try:
            self._write_pkg_info(tmpdir)
            # we expect _find_pkg_info() to find the file in the tmpdir
            self.assertEqual(
                renderspec.utils._find_pkg_info(tmpdir),
                os.path.join(tmpdir, 'PKG-INFO')
            )
        finally:
            shutil.rmtree(tmpdir)

    def test__find_pkg_info_not_found(self):
        tmpdir = tempfile.mkdtemp(prefix='renderspec-test_')
        try:
            self.assertEqual(
                renderspec.utils._find_pkg_info(tmpdir),
                None
            )
        finally:
            shutil.rmtree(tmpdir)

    def test__version_from_pkg_info(self):
        tmpdir = tempfile.mkdtemp(prefix='renderspec-test_')
        version = '5.10.0'
        try:
            self._write_pkg_info(tmpdir, version)
            pkg_info_file = renderspec.utils._find_pkg_info(tmpdir)
            self.assertEqual(
                renderspec.utils._get_version_from_pkg_info(pkg_info_file),
                version
            )
        finally:
            shutil.rmtree(tmpdir)

    @data(
        (['foo-1.2.3.tar.gz'], 'foo', ['foo-1.2.3.tar.gz']),
        (['foo-1.2.3.tar.gz', 'bar-1.2.3.xz'], 'foo', ['foo-1.2.3.tar.gz']),
        # now 2 valid archives - latest created one should be first
        (['foo-1.2.3.tar.gz', 'foo-2.3.4.xz'], 'foo',
         ['foo-2.3.4.xz', 'foo-1.2.3.tar.gz']),
        (['foo-1.2.3.tar.gz'], 'bar', []),
    )
    @unpack
    def test__find_archives(self, archives, pypi_name, expected):
        tmpdir = tempfile.mkdtemp(prefix='renderspec-test_')
        expected = [os.path.join(tmpdir, e) for e in expected]
        try:
            for a in archives:
                open(os.path.join(tmpdir, a), 'w').close()
                time.sleep(0.1)
            self.assertEqual(
                renderspec.utils._find_archives(tmpdir, pypi_name),
                expected
            )
        finally:
            shutil.rmtree(tmpdir)

    def test__find_archives_multiple_dirs(self):
        tmpdir1 = tempfile.mkdtemp(prefix='renderspec-test_')
        tmpdir2 = tempfile.mkdtemp(prefix='renderspec-test_')
        try:
            open(os.path.join(tmpdir2, 'foo-1.2.3.tar.xz'), 'w').close()
            self.assertEqual(
                renderspec.utils._find_archives([None, tmpdir1, tmpdir2],
                                                'foo'),
                [os.path.join(tmpdir2, 'foo-1.2.3.tar.xz')]
            )
        finally:
            shutil.rmtree(tmpdir1)
            shutil.rmtree(tmpdir2)

    def test__find_archives_only_no_dir(self):
        self.assertEqual(renderspec.utils._find_archives([None], 'foo'), [])


if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
# Copyright (c) 2016 SUSE Linux GmbH
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import print_function

from packaging.requirements import Requirement
from packaging.version import Version


def get_requirements(lines):
    """parse the given lines and return a dict with pkg_name->version.
    lines must follow PEP0508"""
    requires = {}
    for l in lines:
        # skip comments and empty lines
        if l.startswith('#') or len(l.strip()) == 0:
            continue
        # remove trailing comments
        l = l.split('#')[0].rstrip(' ')
        r = Requirement(l)
        # check if we need the requirement
        if r.marker:
            # TODO (toabctl): currently we hardcode python 2.7 and linux2
            # see https://www.python.org/dev/peps/pep-0508/#environment-markers
            marker_env = {'python_version': '2.7', 'sys_platform': 'linux'}
            if not r.marker.evaluate(environment=marker_env):
                continue
        if r.specifier:
            # we want the lowest possible version
            # NOTE(toabctl): "min(r.specifier)" doesn't work.
            # see https://github.com/pypa/packaging/issues/69
            lowest = None
            for s in r.specifier:
                # we don't want a lowest version which is not allowed
                if s.operator == '!=':
                    continue
                if not lowest or Version(s.version) < lowest:
                    lowest = Version(s.version)

            if lowest:
                requires[r.name] = str(lowest)
    return requires
